/*******************************************
 * Setup code to collect OV7670 data using
 * CSI hardware and DMA.  It can run in the backgroud
 * and the camera's RGB565 data appears 'magically' in the
 * assigned data buffers.   This initial version does
 * full VGA resolution at up to 30FPS.  Since each pixel
 * of RGB565 data requires two bytes, a full frame of VGA
 * requires 640x480x2 bytes, or 614400 bytes.  The SCI can  
 * alternate between two buffers. Since a 614KB
 * buffer is too large for a single segment of the T4.1 internal
 * RAM, you will need to have the 8MB PSRAM installed.
 * 
 * MJB  10/26/2020
 */

/**************************************************
*  These defines aren't used, but they're a handy
*  reminder of how the )V7670 gets wired to the T4.1
#define RST_PIN 14 // hardware reset for my board

#define SCL_PIN 19
#define SDA_PIN 18

#define ECLK_PIN 41// CSI_MCLK AD_B1_05
#define VD_PIN 17 //CSI_VSYNC AD_B1_06
#define HD_PIN 16 // CSI_HSYNC AD_B1_07
#define DCLK_PIN 40 // CSI_PIXCLK AD_B1_04

#define D0_PIN 27// AD_B1_15
#define D1_PIN 26// AD_B1_14
#define D2_PIN 39// AD_B1_13
#define D3_PIN 38// AD_B1_12
#define D4_PIN 21// AD_B1_11
#define D5_PIN 20// AD_B1_10
#define D6_PIN 23// AD_B1_09
#define D7_PIN 22// AD_B1_08
**********************************************/
#ifndef _CSI_41_H_
#define _CSI_41_H_
#include <Arduino.h>
#include <Wire.h>

// define the bits in the CSISR status register
#define BIT_SOF_INT 16
#define BIT_EOF_INT 17
extern uint32_t last_fc;
extern uint32_t csisr_state;

extern void CSI_Start(void);

extern void CSI_Stop(void);
extern void CSIService(void);
extern void CSI_InitMCLK(uint16_t MHz);
extern void CSI_Setup(uint16_t fwidth, uint16_t fheight, uint8_t *fb1, uint8_t *fb2);
extern void CSI_InitIOMux(void);
extern void print_csi_registers();

// Defines in OV7670_MB.cpp

// Read a single uint8_t from address and return it as a uint8_t
extern uint8_t cameraReadRegister(uint8_t reg);

// Writes a single uint8_t (data) into address
extern void cameraWriteRegister(uint8_t reg, uint8_t data);
extern void cameraWriteRegList(struct regval_entry reglist[]);
extern void cameraReset();
extern void cameraReadAll(uint8_t rvals[]);
extern void cameraShowAll(uint8_t rvals[]);
extern void cameraBegin(uint16_t camwidth, uint16_t camheight, uint8_t *fb1, uint8_t *fb2);
extern void cameraClock(int mhz); // Gets XCLK from masterfreq in CSI_41
extern bool cameraReady();

// Stuff from main ino file
#define FRAME_WIDTH 640l
#define FRAME_HEIGHT 480l

#define FB_WIDTH 640l
#define FB_HEIGHT 480l
#define FB_COUNT 2  // two frames in EXTMEM

#define FRAMEBYTES (FRAME_WIDTH * FRAME_HEIGHT *2)
#define FRAMEBUFFBYTES (FB_WIDTH *FB_HEIGHT *2)

extern uint8_t fb1[FRAMEBUFFBYTES] EXTMEM;
extern uint8_t fb2[FRAMEBUFFBYTES] EXTMEM;

extern uint32_t masterfreq;
extern volatile uint32_t last_csistat, csiIrqCount;
extern volatile uint32_t fb1count, fb2count;
extern volatile uint32_t bswitcherrors;
extern uint32_t framecount;

//const int ledpin    = 13;
extern const int imarkpin;
extern const int pinCamReset;


//#define  LEDON digitalWriteFast(ledpin, HIGH); // Also marks IRQ handler timing
//#define  LEDOFF digitalWriteFast(ledpin, LOW);
//#define  LEDTOGGLE digitalToggleFast(ledpin);

#define  IMARKHI digitalWriteFast(imarkpin, HIGH); // Also marks IRQ handler timing
#define  IMARKLO digitalWriteFast(imarkpin, LOW);
#define  IMARKTOGGLE digitalToggleFast(imarkpin);



#endif //_CSI_41_H_