
//  t4.1 PIN input  and image packet test
//  New CameraBB with new pinouts for wired breadboard
//  MJB  10/27/2020

#include <stdint.h>
#include <Wire.h>
#include <ILI9341_t3n.h>
#include "CSI_41.h"

//#define FRAME_WIDTH 640l
//#define FRAME_HEIGHT 480l

//#define FB_WIDTH 640l
//#define FB_HEIGHT 480l
//#define FB_COUNT 2  // two frames in EXTMEM

//#define FRAMEBYTES (FRAME_WIDTH * FRAME_HEIGHT *2)
//#define FRAMEBUFFBYTES (FB_WIDTH *FB_HEIGHT *2)

uint8_t fb1[FRAMEBUFFBYTES] EXTMEM;
uint8_t fb2[FRAMEBUFFBYTES] EXTMEM;

uint32_t masterfreq = 24;
volatile uint32_t last_csistat, csiIrqCount;
volatile uint32_t fb1count, fb2count;
volatile uint32_t bswitcherrors;
uint32_t framecount;

//ILI9341
#define TFT_CS   10  // AD_B0_02
#define TFT_DC   9  // AD_B0_03
#define TFT_RST  8
ILI9341_t3n tft = ILI9341_t3n(TFT_CS, TFT_DC, TFT_RST);
uint16_t image[320 * 240] DMAMEM;

extern "C" {
  extern  uint8_t external_psram_size;
}



const char compileTime [] = " Compiled on " __DATE__ " " __TIME__;

//const int ledpin    = 13;
const int imarkpin = 32;
const int pinCamReset = 14;


//#define  LEDON digitalWriteFast(ledpin, HIGH); // Also marks IRQ handler timing
//#define  LEDOFF digitalWriteFast(ledpin, LOW);
//#define  LEDTOGGLE digitalToggleFast(ledpin);

//#define  IMARKHI digitalWriteFast(imarkpin, HIGH); // Also marks IRQ handler timing
//#define  IMARKLO digitalWriteFast(imarkpin, LOW);
//#define  IMARKTOGGLE digitalToggleFast(imarkpin);

void setup() {
  while (!Serial && millis() < 4000) ;
  Serial.begin(9600);
  Wire.begin();

  delay(100);
  //pinMode(ledpin, OUTPUT);
  pinMode(imarkpin, OUTPUT);
  pinMode(pinCamReset, OUTPUT);

  delay(10);
  memset(&fb1, 0, FRAMEBUFFBYTES * 2);

  digitalWriteFast(pinCamReset, LOW);
  delay(10);
  digitalWriteFast(pinCamReset, HIGH);  // subsequent resets via SCB
//  delay(10);
//  digitalWriteFast(pinCamReset, LOW);  // subsequent resets via SCB


  Serial.printf("\n\nOV7670 Camera Test 3 %s\n", compileTime);
  Serial.printf("EXT Memory size: %u\n", external_psram_size);
  cameraBegin(FRAME_WIDTH, FRAME_HEIGHT, fb1, fb2);
  // Start ILI9341
  tft.begin();
  tft.setRotation(1);
  tft.fillScreen(ILI9341_RED);
  delay(500);
  tft.fillScreen(ILI9341_GREEN);
  delay(500);
  tft.fillScreen(ILI9341_BLUE);
  delay(500);
  tft.fillScreen(ILI9341_BLACK);
  Serial.println("After cameraBegin(end setup)");
  Serial.println("s - Show Information");
  Serial.println("c - Print CSI Registers");
  Serial.println("r - Show Camera registers");
  Serial.println("f - Show a frame");


}

void loop() {
  static uint32_t lastfc;
  static uint32_t fc = 0;
  char ch;
  if (Serial.available()) {
    
    ch = Serial.read();
    Serial.printf("Command character received: %c\n", ch);
    if (ch == 's') CMSI();
    if (ch == 'c') CMCS();
    if (ch == 'r') CMCR();
    if (ch == 'f') CMGF();
  }

  //  This extends frame count from 16 to 32 bits
  fc  = CSI_CSICR3 >> 16; // get the 16-bit frame count

  framecount += (fc - lastfc);
  lastfc = fc;
}

// Show information
void CMSI(void) {

  Serial.printf("\n\nOV7670 Camera Test 3 %s\n", compileTime);
  Serial.printf("buff1 at %p   buff2 at %p\n", &fb1, &fb2);
  Serial.printf("Camera frame count = %lu \n", framecount);
  Serial.printf("Base Address Change errors: %lu\n", bswitcherrors);
  Serial.printf("csistat: %08X, csiIrqCount: %lu\n", CSI_CSISR, csiIrqCount);
  Serial.printf("fb1 count: %6lu   fb2 count %6lu\n", fb1count, fb2count);
}

void CMCS(void) {  // Show CSI registers
  print_csi_registers();
}



// Show Camera Registers
void CMCR(void) {
  uint8_t regs[200];
  cameraReadAll(regs);
  cameraShowAll(regs);
}


// Get a frame of data
void CMGF(void) {
  uint32_t imagesize, fbc;
  imagesize = (FRAME_WIDTH * FRAME_HEIGHT * 2);
  fbc = fb2count;
  Serial.println("Waiting for next FB2");
  while (fbc == fb2count); // wait until a new frame in fb2 is finished
  Serial.printf("FB2: %lu\n", fb2count);

  arm_dcache_delete(fb2, imagesize);

  // First lets dump the first 256 bytes to see if it makes sense. 
  Serial.println("\n*** Dump first 256 bytes ***");
  for (uint16_t i=0; i < 256; i++) {
    Serial.printf("%02x ", fb2[i]);
    if ((i & 0x1f) == 0x1f ) Serial.println();
  }
  Serial.println("\n*** Dump start at middle 256 bytes ***");
  for (uint16_t i=0; i < 256; i++) {
    if ((i & 0x1f) == 0 ) Serial.printf("%08x: ", (uint32_t) &fb2[FRAME_WIDTH * FRAME_HEIGHT + i]);
    Serial.printf("%02x ", fb2[FRAME_WIDTH * FRAME_HEIGHT + i]);
    if ((i & 0x1f) == 0x1f ) Serial.println();
  }


  // Lets extract image from center.
  // hard code first pass:
  Serial.println("\n*** Dump center colors");
  uint16_t *pimage = image;  
  uint16_t image_offset_y = (FRAME_HEIGHT - tft.height()) / 2;
  uint16_t image_offset_x = (FRAME_WIDTH - tft.width()) / 2;
  uint32_t tft_image_offset = ((image_offset_y * FRAME_WIDTH) + image_offset_x) * 2;
  Serial.printf("Image offsets: X:%u Y: %u bytes: %lu Total: %lu\n", image_offset_x, image_offset_y, tft_image_offset, imagesize);

  uint16_t *pfb = (uint16_t*)((uint8_t*)(fb2 + tft_image_offset));  // Point to pixels.
  uint16_t *pfb_row = pfb;
  for (int row = 0; row < (int)tft.height(); row++) {
    pfb = pfb_row;
    if ((row > 110) && (row < 130)) Serial.printf("%08x: ", (uint32_t)pfb);
    for (int col = 0; col < (int)tft.width(); col++) {
      *pimage = *pfb++;
      // Lets get an idea of data coming in.  How about center 16x16 pixels.
      if ((row > 110) && (row < 130) && (col > 150) && (col < 170)) Serial.printf(" %04x", *pimage);
      pimage++;
    }
    pfb_row += FRAME_WIDTH;
      if ((row > 110) && (row < 130)) Serial.println();
  }

  tft.writeRect(0, 0, tft.width(), tft.height(), image);

}
