//  OV7670 driver using new pinouts for wired breadboard and CSI
//  This version assumes that CSI_41.ino is in the sketch folder
//
#include "CSI_41.h"
#include "OV7670_RegDefs.h"

/*
  const int PinCamVsync = 17; //
  const int PinCamHref  = 16; //
  const int PinCamPpclk = 40; //
  const int PinCamXclk  = 41; //
  const int PinCamD0    = 27; // AD_B1_15
  const int PinCamD1    = 26; //
  const int PinCamD2    = 39; //
  const int PinCamD3    = 38; //
  const int PinCamD4    = 21; //
  const int PinCamD5    = 20; //
  const int PinCamD6    = 23; //
  const int PinCamD7    = 22; // AD_B1_8
*/
// define I2C address for OV7670
static int const camAddress = 0x21; // i2c address of camera's sccb device

struct regval_entry {
  uint8_t reg_num;
  uint8_t value;
};


// Read a single uint8_t from address and return it as a uint8_t
uint8_t cameraReadRegister(uint8_t reg) {
  Wire.beginTransmission(camAddress);
  Wire.write(reg);
  Wire.endTransmission();
  Wire.requestFrom(camAddress, 1);
  uint8_t data = Wire.read();
  Wire.endTransmission();
  return data;
}

// Writes a single uint8_t (data) into address
void cameraWriteRegister(uint8_t reg, uint8_t data) {
  Wire.beginTransmission(camAddress);
  Wire.write(reg); Wire.write(data);
  Wire.endTransmission();
}


// simplified test for end of reglist is  reg_num == 0xFF
// This will have to change if we ever get a camera with
// 255 or more registers!
void cameraWriteRegList(struct regval_entry reglist[]) {
  uint16_t index = 0; 
  uint8_t reg_addr, reg_val;
  while (reglist[index].reg_num != 0xFF) {
    reg_addr = reglist[index].reg_num;
    reg_val = reglist[index].value;
    cameraWriteRegister(reg_addr, reg_val);
    delayMicroseconds(100);
    index++;
  }
}


// Reset all OV7670 registers to their hardware default values
void cameraReset() {
  cameraWriteRegister(0x12, 0x80); delay(10);
  cameraWriteRegister(0x12, 0); delay(10);
}

// Read and display all of the OV7670 registers
// more compact display  MJB 10/3/2020
void cameraReadAll(uint8_t rvals[]) {
  for (int i = 0; i <= 0xC9; i ++) {
    rvals[i] = cameraReadRegister(i);
  }
}

void cameraShowAll(uint8_t rvals[]) {
  for (int i = 0; i <= 0xC9; i ++) {
    if ((i % 8) == 0)Serial.println();
    Serial.printf("%02X:%02X  ", i, rvals[i]);
  }
  Serial.println();
}

struct regval_entry vga_ov7670[]   = {
  {REG_COM14, 0x18},  // divide by 1   add3ed by MJB 10/18/20
  {0x72, 0x0},   // downsample by 1
  {0x73, 0xf0},   // divide by 1
  {0x15, 0x22},  // No PCLK during H blanking negative-going  VSync
  {REG_HREF, 0xB6}, // was 24
  {0x17, 0x16},   // HSTART
  {0x18, 0x04},   // HSTOP  was 04
  {0x19, 0x01},   // VSTARTwas 02
  {0x1a, 0x79},   // VSTOP  was 7A
  {REG_VREF, 0x3F}, // VREF was 0A
  {0xff, 0xff},   /* END MARKER */
};

struct regval_entry rgb565_ov7670[]   = {
  {REG_COM7, COM7_RGB}, /* Selects RGB mode */
  {REG_RGB444, 0},    /* No RGB444 please */
  {REG_COM1, 0x0},
  {REG_COM15, 0xD0},
  {REG_COM9, 0x6A},  /* 128x gain ceiling; 0x8 is reserved bit */
  {0x4f, 0xb3},    /* "matrix coefficient 1" */
  {0x50, 0xb3},    /* "matrix coefficient 2" */
  {0x51, 0},     /* vb */
  {0x52, 0x3d},    /* "matrix coefficient 4" */
  {0x53, 0xa7},    /* "matrix coefficient 5" */
  {0x54, 0xe4},    /* "matrix coefficient 6" */
  {REG_COM13, /*COM13_GAMMA|*/COM13_UVSAT},

  // added a boost in contrast and brightness MJB 10/27/2020
  {REG_BRIGHT, 0X32},  // default is 0
  {REG_CONTRAS,0x5C},  // default is 0x40
  
  {0xff, 0xff}, /* END MARKER */
};


struct regval_entry ov7670_default_regs[]   = {//from the linux driver
  {REG_COM7, COM7_RESET},
  {REG_TSLB,  0x04},  /* OV */
  {REG_COM7, 0},  /* VGA */
  /*
     Set the hardware window.  These values from OV don't entirely
     make sense - hstop is less than hstart.  But they work...
  */
  {REG_HSTART, 0x16}, {REG_HSTOP, 0x04},
  {REG_HREF, 0xB6}, {REG_VSTART, 0x01},
  {REG_VSTOP, 0x79},  {REG_VREF, 0x3F},

  {REG_COM3, 0},  {REG_COM14, 0},
  /* Mystery scaling numbers */
  {0x70, 0x3a},   {0x71, 0x35},
  {0x72, 0x11},   {0x73, 0xf0},
  {0xa2,/* 0x02 changed to 1*/1}, {REG_COM10, COM10_VS_NEG},
  /* Gamma curve values */

  {0x7a, 0x20},   {0x7b, 0x10},
  {0x7c, 0x1e},   {0x7d, 0x35},
  {0x7e, 0x5a},   {0x7f, 0x69},
  {0x80, 0x76},   {0x81, 0x80},
  {0x82, 0x88},   {0x83, 0x8f},
  {0x84, 0x96},   {0x85, 0xa3},
  {0x86, 0xaf},   {0x87, 0xc4},
  {0x88, 0xd7},   {0x89, 0xe8},

  /* AGC and AEC parameters.  Note we start by disabling those features,
     then turn them only after tweaking the values. */
  {REG_COM8, COM8_FASTAEC | COM8_AECSTEP},
  {REG_GAIN, 0},  {REG_AECH, 0},
  {REG_COM4, 0x40}, /* magic reserved bit */
  {REG_COM9, 0x18}, /* 4x gain + magic rsvd bit */
  {REG_BD50MAX, 0x05},  {REG_BD60MAX, 0x07},
  {REG_AEW, 0x95},  {REG_AEB, 0x33},
  {REG_VPT, 0xe3},  {REG_HAECC1, 0x78},
  {REG_HAECC2, 0x68}, {0xa1, 0x03}, /* magic */
  {REG_HAECC3, 0xd8}, {REG_HAECC4, 0xd8},
  {REG_HAECC5, 0xf0}, {REG_HAECC6, 0x90},
  {REG_HAECC7, 0x94},
  {REG_COM8, COM8_FASTAEC | COM8_AECSTEP | COM8_AGC | COM8_AEC},
  {0x30, 0}, {0x31, 0}, //disable some delays

  // {0xff, 0xff}, /* END MARKER  for testing */

  /* Almost all of these are magic "reserved" values.  */
  {REG_COM5, 0x61}, {REG_COM6, 0x4b},
  {0x16, 0x02},   {REG_MVFP, 0x07},
  {0x21, 0x02},   {0x22, 0x91},
  {0x29, 0x07},   {0x33, 0x0b},
  {0x35, 0x0b},   {0x37, 0x1d},
  {0x38, 0x71},   {0x39, 0x2a},
  {REG_COM12, 0x78},  {0x4d, 0x40},
  {0x4e, 0x20},   {REG_GFIX, 0},
  /*{0x6b, 0x4a},*/   {0x74, 0x10},
  {0x8d, 0x4f},   {0x8e, 0},
  {0x8f, 0},    {0x90, 0},
  {0x91, 0},    {0x96, 0},
  {0x9a, 0},    {0xb0, 0x84},
  {0xb1, 0x0c},   {0xb2, 0x0e},
  {0xb3, 0x82},   {0xb8, 0x0a},

  /* More reserved magic, some of which tweaks white balance */
  {0x43, 0x0a},   {0x44, 0xf0},
  {0x45, 0x34},   {0x46, 0x58},
  {0x47, 0x28},   {0x48, 0x3a},
  {0x59, 0x88},   {0x5a, 0x88},
  {0x5b, 0x44},   {0x5c, 0x67},
  {0x5d, 0x49},   {0x5e, 0x0e},
  {0x6c, 0x0a},   {0x6d, 0x55},
  {0x6e, 0x11},   {0x6f, 0x9e}, /* it was 0x9F "9e for advance AWB" */
  {0x6a, 0x40},   {REG_BLUE, 0x40},
  {REG_RED, 0x60},
  {REG_COM8, COM8_FASTAEC | COM8_AECSTEP | COM8_AGC | COM8_AEC | COM8_AWB},

  /* Matrix coefficients */
  {0x4f, 0x80},   {0x50, 0x80},
  {0x51, 0},    {0x52, 0x22},
  {0x53, 0x5e},   {0x54, 0x80},
  {0x58, 0x9e},

  {REG_COM16, COM16_AWBGAIN}, {REG_EDGE, 0},
  {0x75, 0x05},   {REG_REG76, 0xe1},
  {0x4c, 0},    {0x77, 0x01},
  {REG_COM13, /*0xc3*/0x48},  {0x4b, 0x09},
  {0xc9, 0x60},   /*{REG_COM16, 0x38},*/
  {0x56, 0x40},

  {0x34, 0x11},   {REG_COM11, COM11_EXP | COM11_HZAUTO},
  {0xa4, 0x82/*Was 0x88*/},   {0x96, 0},
  {0x97, 0x30},   {0x98, 0x20},
  {0x99, 0x30},   {0x9a, 0x84},
  {0x9b, 0x29},   {0x9c, 0x03},
  {0x9d, 0x4c},   {0x9e, 0x3f},
  {0x78, 0x04},

  /* Extra-weird stuff.  Some sort of multiplexor register */
  {0x79, 0x01},   {0xc8, 0xf0},
  {0x79, 0x0f},   {0xc8, 0x00},
  {0x79, 0x10},   {0xc8, 0x7e},
  {0x79, 0x0a},   {0xc8, 0x80},
  {0x79, 0x0b},   {0xc8, 0x01},
  {0x79, 0x0c},   {0xc8, 0x0f},
  {0x79, 0x0d},   {0xc8, 0x20},
  {0x79, 0x09},   {0xc8, 0x80},
  {0x79, 0x02},   {0xc8, 0xc0},
  {0x79, 0x03},   {0xc8, 0x40},
  {0x79, 0x05},   {0xc8, 0x30},
  {0x79, 0x26},

  {0xff, 0xff}, /* END MARKER */
};

#define DEFAULTCLOCK 12
// Start up the camera.  Assume that MCLK is already present
void cameraBegin(uint16_t camwidth, uint16_t camheight, uint8_t *fb1, uint8_t *fb2) {
  CSI_InitIOMux(); // set up connections to IO Pins
  CSI_InitMCLK(DEFAULTCLOCK);
  cameraClock(DEFAULTCLOCK);  // default clock
  delay(150);
  cameraWriteRegList(ov7670_default_regs);
  delay(10);
  // cameraWriteRegList(vga_ov7670);
  delay(10);
  cameraWriteRegList(rgb565_ov7670);

  CSI_Stop();
  CSI_Setup(camwidth, camheight, fb1, fb2);
  CSI_Start();
}



void cameraClock(int mhz) // Gets XCLK from masterfreq in CSI_41
{ // Clock = external-clock * pll_mull(0,4,6 or 8 on bits 6 and 7) / (2*divider_5_lsb+1)
  float r, ratio, bestr;
  uint8_t reg;
  uint16_t bestpll, bestdiv, p, d;
  float const pll[3] = { 4.0f, 6.0f, 8.0f };
  ratio = (float) mhz / (float)masterfreq; // this is what we want to get to... find closest value...

  bestpll = 0, bestdiv = 0; bestr = 1000.0;
  for (p = 0; p < 3; p++) {
    for (d = 0; d < 32; d++) {
      r = pll[p] / (2 * (d + 1));
      if (fabs(r - ratio) < fabs(ratio - bestr)) {
        bestr = r;
        bestpll = p + 1;
        bestdiv = d;
      }
    }
  }
  // CLKRC register: Prescaler divide by 31
  reg = cameraReadRegister(0x11); cameraWriteRegister(0x11, (reg & 0b1000000) | bestdiv);
  // DBLV register: PLL = 6
  reg = cameraReadRegister(0x6B); cameraWriteRegister(0x6B, (reg & 0b00111111) | (bestpll << 6));
  Serial.printf( "cam clock %6.2f with pll=%d and div=%d\n", masterfreq * bestr, bestpll, bestdiv);

}

// return true if a new frame is available in the cameraImage buffer.
bool cameraReady() {
  static  uint32_t lastfc;
  bool r;
  r = (framecount != lastfc);
  lastfc = framecount;
  return r;
}
